/*	EFUNC.H:	MicroEMACS function declarations and names

		This file list all the C code functions used by MicroEMACS
	and the names to use to bind keys to them. To add functions,
	declare it here in both the extern function list and the name
	binding table.

	Update History:

	Daniel Lawrence
		29-jan-86
		- separeted out external declarations to a separate .h file
		- built original name to function binding table
		30-jan-86
		- added command declarations for Fill Paragraph command
		11-feb-86
		- added declaration for help and describe key commands
		13-feb-86
		- added declarations for view-file command
		15-feb-86
		- added declataitons for insert file command
		17-feb-86
		- added declarations for scroll next window up/down
		20-feb-86
		- expanded binding table to make room for new bindings
		24-feb-86
		- added declarations for bind-to-key and unbind-key
		  execute named command and describe bindings
		26-feb-86
		- added declarations for execute command
		- changed name of "visit-file" to "find-file"
		- added declaration for execute-buffer command
		27-feb-86
		- added declarations for execute-file command
		3-apr-86
		- added declarations for next-buffer command
		5-apr-86
		- added declarations for kill-paragraph command
		8-apr-86
		- added declarations for add/delete global mode
		8-apr-86
		- added declarations for insert space command
*/

/*	External function declarations		*/

extern  int     ctrlg(int f, int n);    /* Abort out of things          */
extern  int     quit(int f, int n);     /* Quit                         */
extern  int     ctlxlp(int f, int n);   /* Begin macro                  */
extern  int     ctlxrp(int f, int n);   /* End macro                    */
extern  int     ctlxe(int f, int n);                /* Execute macro                */
extern  int     fileread(int f, int n);             /* Get a file, read only        */
extern  int     filefind(int f, int n);		/* Get a file, read write       */
extern  int     filewrite(int f, int n);            /* Write a file                 */
extern  int     filesave(int f, int n);             /* Save current file            */
extern  int     filename(int f, int n);             /* Adjust file name             */
extern  int     getccol(int bflg);              /* Get current column           */
extern  int     gotobol(int f, int n);              /* Move to start of line        */
extern  int     forwchar(int f, int n);             /* Move forward by characters   */
extern  int     gotoeol(int f, int n);              /* Move to end of line          */
extern  int     backchar(int f, int n);             /* Move backward by characters  */
extern  int     forwline(int f, int n);             /* Move forward by lines        */
extern  int     backline(int f, int n);             /* Move backward by lines       */
extern  int     forwpage(int f, int n);             /* Move forward by pages        */
extern  int     backpage(int f, int n);             /* Move backward by pages       */
extern  int     gotobob(int f, int n);              /* Move to start of buffer      */
extern  int     gotoeob(int f, int n);              /* Move to end of buffer        */
extern  int     setfillcol(int f, int n);           /* Set fill column.             */
extern  int     setmark(int f, int n);              /* Set mark                     */
extern  int     swapmark(int f, int n);             /* Swap "." and mark            */
extern  int     forwsearch(int f, int n);           /* Search forward               */
extern  int     backsearch(int f, int n);           /* Search backwards             */
extern	int	sreplace(int f, int n);		/* search and replace		*/
extern	int	qreplace(int f, int n);		/* search and replace w/query	*/
extern  int     showcpos(int f, int n);             /* Show the cursor position     */
extern  int     nextwind(int f, int n);             /* Move to the next window      */
extern  int     prevwind(int f, int n);             /* Move to the previous window  */
extern  int     onlywind(int f, int n);             /* Make current window only one */
extern  int     splitwind(int f, int n);            /* Split current window         */
extern  int     mvdnwind(int f, int n);             /* Move window down             */
extern  int     mvupwind(int f, int n);             /* Move window up               */
extern  int     enlargewind(int f, int n);          /* Enlarge display window.      */
extern  int     shrinkwind(int f, int n);           /* Shrink window.               */
extern  int     listbuffers(int f, int n);          /* Display list of buffers      */
extern  int     usebuffer(int f, int n);            /* Switch a window to a buffer  */
extern  int     killbuffer(int f, int n);           /* Make a buffer go away.       */
extern  int     reposition(int f, int n);           /* Reposition window            */
extern  int     refresh(int f, int n);              /* Refresh the screen           */
extern  int     twiddle(int f, int n);              /* Twiddle characters           */
extern  int     tab(int f, int n);                  /* Insert tab                   */
extern  int     newline(int f, int n);              /* Insert CR-LF                 */
extern  int     indent(int f, int n);               /* Insert CR-LF, then indent    */
extern  int     openline(int f, int n);             /* Open up a blank line         */
extern  int     deblank(int f, int n);              /* Delete blank lines           */
extern  int     quote(int f, int n);                /* Insert literal               */
extern  int     backword(int f, int n);             /* Backup by words              */
extern  int     forwword(int f, int n);             /* Advance by words             */
extern  int     forwdel(int f, int n);              /* Forward delete               */
extern  int     backdel(int f, int n);              /* Backward delete              */
extern  int     killtext(int f, int n);             /* Kill forward                 */
extern  int     yank(int f, int n);                 /* Yank back from killbuffer.   */
extern  int     upperword(int f, int n);            /* Upper case word.             */
extern  int     lowerword(int f, int n);            /* Lower case word.             */
extern  int     upperregion(int f, int n);          /* Upper case region.           */
extern  int     lowerregion(int f, int n);          /* Lower case region.           */
extern  int     capword(int f, int n);              /* Initial capitalize word.     */
extern  int     delfword(int f, int n);             /* Delete forward word.         */
extern  int     delbword(int f, int n);             /* Delete backward word.        */
extern  int     killregion(int f, int n);           /* Kill region.                 */
extern  int     copyregion(int f, int n);           /* Copy region to kill buffer.  */
extern  int     spawncli(int f, int n);             /* Run CLI in a subjob.         */
extern  int     spawn(int f, int n);                /* Run a command in a subjob.   */
extern  int     quickexit(int f, int n);            /* low keystroke style exit.    */
extern	int	setmode(int f, int n);		/* set an editor mode		*/
extern	int	delmode(int f, int n);		/* delete a mode		*/
extern	int	gotoline(int f, int n);		/* go to a numbered line	*/
extern	int	namebuffer(int f, int n);		/* rename the current buffer	*/
extern	int	gotobop(int f, int n);		/* go to begining/paragraph	*/
extern	int	gotoeop(int f, int n);		/* go to end/paragraph		*/
extern	int	fillpara(int f, int n);		/* fill current paragraph	*/
extern	int	help(int f, int n);			/* get the help file here	*/
extern	int	deskey(int f, int n);		/* describe a key's binding	*/
extern	int	viewfile(int f, int n);		/* find a file in view mode	*/
extern	int	insfile(int f, int n);		/* insert a file		*/
extern	int	scrnextup(int f, int n);		/* scroll next window back	*/
extern	int	scrnextdw(int f, int n);		/* scroll next window down	*/
extern	int	bindtokey(int f, int n);		/* bind a function to a key	*/
extern	int	unbindkey(int f, int n);		/* unbind a key's function	*/
extern	int	namedcmd(int f, int n);		/* execute named command	*/
extern	int	desbind(int f, int n);		/* describe bindings		*/
extern	int	execcmd(int f, int n);		/* execute a command line	*/
extern	int	execbuf(int f, int n);		/* exec commands from a buffer	*/
extern	int	execfile(int f, int n);		/* exec commands from a file	*/
extern	int	nextbuffer(int f, int n);		/* switch to the next buffer	*/
extern	int	killpara(int f, int n);		/* kill the current paragraph	*/
extern	int	setgmode(int f, int n);		/* set a global mode		*/
extern	int	delgmode(int f, int n);		/* delete a global mode		*/
extern	int	insspace(int f, int n);		/* insert a space forword	*/
extern	int	forwhunt(int f, int n);		/* hunt forward for next match	*/
extern	int	backhunt(int f, int n);		/* hunt backwards for next match*/

extern int mlreply(char *prompt, char *buf, int nbuf);
extern int mlwrite(char *x, ...);
extern int mlerase(void);
extern int mlputs(char *s);
extern int inword(void);
extern int ffropen(char *fn);
extern int ffclose(void);
extern int getfile(char fname[], int lockfl);
extern int (*getname(void))(int f, int n);
extern int bclear(register BUFFER *bp);
extern int linsert(int n, int c);
extern int lnewline(void);
extern int (*fncmatch(char *fname))(int f, int n);
extern int getkey(void);
extern int getctl(void);
extern int makename(char bname[], char fname[]);
extern int readin(char fname[], int lockfl);
extern int zotbuf(register BUFFER *bp);
extern int swbuffer(BUFFER *bp);
extern int mlyesno(char *prompt);
extern int lfree(register LINE *lp);
extern int vtinit(void);
extern int vttidy(void);
extern int movecursor(int row, int col);
extern int typahead(void);
extern int upmode(void);
extern int mlreplyt(char *prompt, char *buf, int nbuf, char eolchar);
extern int nxtarg(char *tok);
extern int rdonly(void);
extern int ffgetline(register char buf[], int nbuf);
extern int ffwopen(char *fn);
extern int ffputline(char buf[], int nbuf);
extern int kinsert(int c);
extern int startup(void);
extern int update(void);
extern int wrapword(void);
extern int ldelete(int n, int kflag);
extern int insbrace(int n, int c);
extern int inspound(void);
extern int anycb(void);
extern int lchange(register int flag);
extern int kdelete(void);
extern int kremove(int n);

#if	V7 & BSD
extern	int	bktoshell();		/* suspend emacs to parent shell*/
extern	int	rtfrmshell();		/* return from a suspended state*/
#endif

/*	Name to function binding table

		This table gives the names of all the bindable functions
	end their C function address. These are used for the bind-to-key
	function.
*/

#ifdef NEEDNAMES

NBIND	names[] = {
	{"add-mode",			setmode},
	{"add-global-mode",		setgmode},
	{"backward-character",		backchar},
	{"begin-macro",			ctlxlp},
	{"begining-of-file",		gotobob},
	{"begining-of-line",		gotobol},
	{"bind-to-key",			bindtokey},
	{"buffer-position",		showcpos},
	{"case-region-lower",		lowerregion},
	{"case-region-upper",		upperregion},
	{"case-word-capitalize",	capword},
	{"case-word-lower",		lowerword},
	{"case-word-upper",		upperword},
	{"change-file-name",		filename},
	{"clear-and-redraw",		refresh},
	{"copy-region",			copyregion},
	{"delete-blank-lines",		deblank},
	{"delete-buffer",		killbuffer},
	{"delete-mode",			delmode},
	{"delete-global-mode",		delgmode},
	{"delete-next-character",	forwdel},
	{"delete-next-word",		delfword},
	{"delete-other-windows",	onlywind},
	{"delete-previous-character",	backdel},
	{"delete-previous-word",	delbword},
	{"describe-bindings",		desbind},
	{"describe-key",		deskey},
	{"end-macro",			ctlxrp},
	{"end-of-file",			gotoeob},
	{"end-of-line",			gotoeol},
	{"exchange-point-and-mark",	swapmark},
	{"execute-buffer",		execbuf},
	{"execute-command-line",	execcmd},
	{"execute-file",		execfile},
	{"execute-macro",		ctlxe},
	{"execute-named-command",	namedcmd},
	{"exit-emacs",			quit},
	{"fill-paragraph",		fillpara},
	{"find-file",			filefind},
	{"forward-character",		forwchar},
	{"goto-line",			gotoline},
	{"grow-window",			enlargewind},
	{"handle-tab",			tab},
	{"hunt-forward",		forwhunt},
	{"hunt-backward",		backhunt},
	{"help",			help},
	{"i-shell",			spawncli},
	{"insert-file",			insfile},
	{"insert-space",		insspace},
	{"kill-paragraph",		killpara},
	{"kill-region",			killregion},
	{"kill-to-end-of-line",		killtext},
	{"list-buffers",		listbuffers},
	{"move-window-down",		mvdnwind},
	{"move-window-up",		mvupwind},
	{"name-buffer",			namebuffer},
	{"newline",			newline},
	{"newline-and-indent",		indent},
	{"next-buffer",			nextbuffer},
	{"next-line",			forwline},
	{"next-page",			forwpage},
	{"next-paragraph",		gotoeop},
	{"next-window",			nextwind},
	{"next-word",			forwword},
	{"open-line",			openline},
	{"previous-line",		backline},
	{"previous-page",		backpage},
	{"previous-paragraph",		gotobop},
	{"previous-window",		prevwind},
	{"previous-word",		backword},
	{"query-replace-string",	qreplace},
	{"quick-exit",			quickexit},
	{"quote-character",		quote},
	{"read-file",			fileread},
	{"redraw-display",		reposition},
	{"replace-string",		sreplace},
	{"save-file",			filesave},
	{"scroll-next-up",		scrnextup},
	{"scroll-next-down",		scrnextdw},
	{"search-forward",		forwsearch},
	{"search-reverse",		backsearch},
	{"select-buffer",		usebuffer},
	{"set-fill-column",		setfillcol},
	{"set-mark",			setmark},
	{"shell-command",		spawn},
	{"shrink-window",		shrinkwind},
	{"split-current-window",	splitwind},
#if	V7 & BSD
	{"suspend-emacs",		bktoshell},
#endif
	{"transpose-characters",	twiddle},
	{"unbind-key",			unbindkey},
	{"view-file",			viewfile},
	{"write-file",			filewrite},
	{"yank",			yank},

	{"",			NULL}
};

#endif
