/*********************************************************************/
/*                                                                   */
/*  This Program Written by Paul Edwards                             */
/*  Released to the Public Domain                                    */
/*                                                                   */
/*********************************************************************/
/*********************************************************************/
/*                                                                   */
/*  modem - a modem that handles ASCII to EBCDIC, nominally via      */
/*  using ATDE (to dial an EBCDIC host)                              */
/*                                                                   */
/*********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <unistd.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <termios.h>

/* do you plan to use extended (-x) on mfterm? */
#define EXTEND 1

/* will you have two splash screens? */
#define DOUBLE 1

#ifdef EBCDIC
#define CHAR_ESC_CHAR '\x27'
#else
#define CHAR_ESC_CHAR '\x1b'
#endif

#define CHAR_XON_CHAR '\x11'
#define CHAR_XOFF_CHAR '\x13'

static int sockfd;
struct hostent *hep;
struct sockaddr_in sockain;

struct termios tios;

static unsigned char buf[10000];
static int cnt;

/* ASCII to EBCDIC */
static unsigned char a2e[] =
"\x00\x01\x02\x03\x37\x2D\x2E\x2F\x16\x05\x15\x0B\x0C\x0D\x0E\x0F"
"\x10\x11\x12\x13\x3C\x3D\x32\x26\x18\x19\x3F\x27\x1C\x1D\x1E\x1F"
"\x40\x5A\x7F\x7B\x5B\x6C\x50\x7D\x4D\x5D\x5C\x4E\x6B\x60\x4B\x61"
"\xF0\xF1\xF2\xF3\xF4\xF5\xF6\xF7\xF8\xF9\x7A\x5E\x4C\x7E\x6E\x6F"
"\x7C\xC1\xC2\xC3\xC4\xC5\xC6\xC7\xC8\xC9\xD1\xD2\xD3\xD4\xD5\xD6"
"\xD7\xD8\xD9\xE2\xE3\xE4\xE5\xE6\xE7\xE8\xE9\xAD\xE0\xBD\x5F\x6D"
"\x79\x81\x82\x83\x84\x85\x86\x87\x88\x89\x91\x92\x93\x94\x95\x96"
"\x97\x98\x99\xA2\xA3\xA4\xA5\xA6\xA7\xA8\xA9\xC0\x4F\xD0\xA1\x07"
"\x20\x21\x22\x23\x24\x25\x06\x17\x28\x29\x2A\x2B\x2C\x09\x0A\x1B"
"\x30\x31\x1A\x33\x34\x35\x36\x08\x38\x39\x3A\x3B\x04\x14\x3E\xFF"
"\x41\xAA\x4A\xB1\x9F\xB2\x6A\xB5\xBB\xB4\x9A\x8A\xB0\xCA\xAF\xBC"
"\x90\x8F\xEA\xFA\xBE\xA0\xB6\xB3\x9D\xDA\x9B\x8B\xB7\xB8\xB9\xAB"
"\x64\x65\x62\x66\x63\x67\x9E\x68\x74\x71\x72\x73\x78\x75\x76\x77"
"\xAC\x69\xED\xEE\xEB\xEF\xEC\xBF\x80\xFD\xFE\xFB\xFC\xBA\xAE\x59"
"\x44\x45\x42\x46\x43\x47\x9C\x48\x54\x51\x52\x53\x58\x55\x56\x57"
"\x8C\x49\xCD\xCE\xCB\xCF\xCC\xE1\x70\xDD\xDE\xDB\xDC\x8D\x8E\xDF"
;

/* EBCDIC to ASCII*/
static unsigned char e2a[] =
"\x00\x01\x02\x03\x9C\x09\x86\x7F\x97\x8D\x8E\x0B\x0C\x0D\x0E\x0F"
"\x10\x11\x12\x13\x9D\x0A\x08\x87\x18\x19\x92\x8F\x1C\x1D\x1E\x1F"
"\x80\x81\x82\x83\x84\x85\x17\x1B\x88\x89\x8A\x8B\x8C\x05\x06\x07"
"\x90\x91\x16\x93\x94\x95\x96\x04\x98\x99\x9A\x9B\x14\x15\x9E\x1A"
"\x20\xA0\xE2\xE4\xE0\xE1\xE3\xE5\xE7\xF1\xA2\x2E\x3C\x28\x2B\x7C"
"\x26\xE9\xEA\xEB\xE8\xED\xEE\xEF\xEC\xDF\x21\x24\x2A\x29\x3B\x5E"
"\x2D\x2F\xC2\xC4\xC0\xC1\xC3\xC5\xC7\xD1\xA6\x2C\x25\x5F\x3E\x3F"
"\xF8\xC9\xCA\xCB\xC8\xCD\xCE\xCF\xCC\x60\x3A\x23\x40\x27\x3D\x22"
"\xD8\x61\x62\x63\x64\x65\x66\x67\x68\x69\xAB\xBB\xF0\xFD\xFE\xB1"
"\xB0\x6A\x6B\x6C\x6D\x6E\x6F\x70\x71\x72\xAA\xBA\xE6\xB8\xC6\xA4"
"\xB5\x7E\x73\x74\x75\x76\x77\x78\x79\x7A\xA1\xBF\xD0\x5B\xDE\xAE"
"\xAC\xA3\xA5\xB7\xA9\xA7\xB6\xBC\xBD\xBE\xDD\xA8\xAF\x5D\xB4\xD7"
"\x7B\x41\x42\x43\x44\x45\x46\x47\x48\x49\xAD\xF4\xF6\xF2\xF3\xF5"
"\x7D\x4A\x4B\x4C\x4D\x4E\x4F\x50\x51\x52\xB9\xFB\xFC\xF9\xFA\xFF"
"\x5C\xF7\x53\x54\x55\x56\x57\x58\x59\x5A\xB2\xD4\xD6\xD2\xD3\xD5"
"\x30\x31\x32\x33\x34\x35\x36\x37\x38\x39\xB3\xDB\xDC\xD9\xDA\x9F"
;

static int lsock;
static int lsockfd;
struct sockaddr_in lsockain;
struct sockaddr newsocka;
static int newsocklen;

/* have we finished TN3270 negotiation? */
static int negfin = 0;

int main(int argc, char **argv)
{
    char *address;
    unsigned int port;
    char *whofirst = "you";
    int c;
    unsigned int lport = 63000; /* listen on this port */
    int gotxon;

    if (argc != 4)
    {
        printf("usage: modem <IP name/address> <port number> <listen port>\n");
        printf("e.g. modem remote.com 1234 4321\n");
        return (EXIT_FAILURE);
    }
    
    /* Now assume they entered ATDEkerravon.mooo.com:64000 */
    address = argv[1];
    port = atoi(argv[2]);
    printf("will connect to %s %u\n", address, port);

    lport = atoi(argv[3]);
    printf("but first you must connect to me on %u\n", lport);

    lsock = socket(AF_INET, SOCK_STREAM, 0);
    printf("return is %d\n", lsock);
    if (lsock < 0) return (EXIT_FAILURE);

    memset(&lsockain, 0, sizeof lsockain);
    lsockain.sin_family = AF_INET;
    lsockain.sin_addr.s_addr = htonl(INADDR_ANY);
    lsockain.sin_port = htons(lport);
    if (bind(lsock,
             (struct sockaddr *)&lsockain,
             sizeof lsockain)
        < 0) return (EXIT_FAILURE);
    printf("bound\n");

    if (listen(lsock, 64) < 0) return (EXIT_FAILURE);
    printf("heard\n");

    lsockfd = accept(lsock,
                     &newsocka,
                     &newsocklen);
    printf("newsock is %d\n", lsockfd);

    sockfd = socket(AF_INET, SOCK_STREAM, 0);
    printf("return is %d\n", sockfd);
    if (sockfd < 0) return (EXIT_FAILURE);

    hep = gethostbyname(address);
    if (hep == NULL) return (EXIT_FAILURE);
    printf("addr0 is %x\n", hep->h_addr[0]);

    memset(&sockain, 0, sizeof sockain);
    sockain.sin_family = AF_INET;
    memcpy(&sockain.sin_addr.s_addr,
           hep->h_addr,
           sizeof sockain.sin_addr.s_addr);
    sockain.sin_port = htons(port);
    if (connect(sockfd,
                (struct sockaddr *)&sockain,
                sizeof sockain)
        < 0) return (EXIT_FAILURE);
    printf("connected\n");

#if 0
    freopen(NULL, "rb", stdin);
    _setmode(_fileno(stdin), O_BINARY);
#endif

#if 0
    setvbuf(stdin, NULL, _IONBF, 0);

    tcgetattr(0, &tios);
    tios.c_iflag &= ~IXOFF;
    tios.c_iflag &= ~IXON;
    tios.c_lflag &= ~ECHO;
    tios.c_lflag &= ~ICANON;
    tcsetattr(0, TCSANOW, &tios);
#endif

    gotxon = 0;
    if (strcmp(whofirst, "you") == 0) goto you;
    while (1)
    {
        int cnt = 0;
                
        while (1)
        {
            int cnt2;
            unsigned char *p;

            /* we read until we get an XOFF. can't use XON
               because PDOS is expecting that */            
            printf("reading from terminal\n");
            cnt2 = read(lsockfd, buf + cnt, sizeof buf - cnt);
            printf("cnt2 is %d\n", cnt2);
            if (cnt2 <= 0)
            {
                printf("cnt too low - exiting\n");
                close(sockfd);
                close(lsockfd);
                exit(EXIT_FAILURE);
            }
            cnt += cnt2;
recheck:
#if EXTEND

#if DOUBLE
            if ((negfin == 2) && (cnt > 5))
#else
            if (negfin && (cnt > 5))
#endif
            {
                p = memchr(buf + 5, CHAR_XOFF_CHAR, cnt - 5);
            }
#if DOUBLE
            else if (negfin == 2)
#else
            else if (negfin)
#endif
            {
                p = NULL;
            }
            else
#endif
            {
                p = memchr(buf, CHAR_XOFF_CHAR, cnt);
            }
            if (p == NULL) continue;
            cnt2 = cnt - (p - buf);
            cnt = p - buf;
            printf("writing %d bytes\n", cnt);
            write(sockfd, buf, cnt);
#if EXTEND
            if (gotxon)
            {
                /* if z/PDOS just sent an XON, we first
                   get here to ACK the message, and then
                   we need to go back to get the actual
                   input from the user */
                gotxon = 0;
                memmove(buf, buf + cnt + 1, cnt2 - 1);
                cnt = cnt2 - 1;
                goto recheck; /* may already have XOFF at this point */
            }
#endif
            break;
        }
        you:
        while (1)
        {
            cnt = read(sockfd, buf, sizeof buf);
            if (cnt <= 0)
            {
                printf("remote died? %d\n", cnt);
                close(sockfd);
                close(lsockfd);
                exit(0);
            }
            c = buf[0];
            printf("just read %02X\n", c);
            write(lsockfd, buf, cnt);
#if DOUBLE
            if (negfin == 2)
#else
            if (negfin)
#endif
            {
#if EXTEND
                if (memchr(buf + 5, CHAR_XON_CHAR, cnt - 5) != NULL)
#else
                if (memchr(buf, CHAR_XON_CHAR, cnt) != NULL)
#endif
                {
                    printf("got xon, breaking\n");
                    gotxon = 1;
                    break;
                }
#if EXTEND
                /* TN3270E requires ack for every message */
                break;
#endif
            }
            /* this will hopefully catch a splash screen, and
               even if it has an XON in it, it won't interfere
               with the XON logic */
            else if ((buf[0] != 0xff)
                     && !((buf[0] == 0x00) && (cnt <= 9)))
            {
                negfin++;
            }
            else
            {
                break;
            }
        }
    }

    printf("terminating\n");
    close(sockfd);
    close(lsockfd);
    return (0);
}
