/*********************************************************************/
/*                                                                   */
/*  This Program Written by Paul Edwards.                            */
/*  Released to the Public Domain                                    */
/*                                                                   */
/*********************************************************************/
/*********************************************************************/
/*                                                                   */
/*  mvsseq - generate/replace MVS sequence numbers                   */
/*                                                                   */
/*  Start=increment was 00001000 for some editors,                   */
/*  00010000 for Wylbur, but                                         */
/*  since ISPF's use of 79/80 it's 000100vv.                         */
/*                                                                   */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static char buf[1000];

int main(int argc, char **argv)
{
    FILE *fp;
    FILE *fq;
    char *in;
    char *out;
    long curr = 10000;
    long inc = 10000;

    if (argc <= 2)
    {
        printf("usage: mvsseq <in file> <out file> [start num] [increment]\n");
        printf("e.g. mvsseq dd:in dd:out 1000 50\n");
        printf("for stdin/stdout use \"-\"\n");
        printf("default start/increment is 10000\n");
        return (EXIT_FAILURE);
    }

    in = *(argv + 1);
    if (strcmp(in, "-") == 0)
    {
        fp = stdin;
    }
    else
    {
        fp = fopen(in, "r");
        if (fp == NULL)
        {
            printf("can't open %s\n", in);
            return (EXIT_FAILURE);
        }
    }

    out = *(argv + 2);
    if (strcmp(out, "-") == 0)
    {
        fq = stdout;
    }
    else
    {
        fq = fopen(out, "w");
        if (fq == NULL)
        {
            printf("can't open %s\n", out);
            return (EXIT_FAILURE);
        }
    }

    if (argc > 3)
    {
        curr = atol(*(argv + 3));
    }

    if (argc > 4)
    {
        inc = atol(*(argv + 4));
    }

    while (fgets(buf, sizeof buf, fp) != NULL)
    {
        char *p;

        p = strchr(buf, '\n');
        if (p == NULL)
        {
            p = buf + strlen(buf);
        }
        if ((p - buf) > 72)
        {
            p = buf + 72;
        }
        memset(p, ' ', buf + 72 - p);
        sprintf(buf + 72, "%08ld\n", curr);
        curr += inc;
        fputs(buf, fq);
    }

    if (ferror(fp) || ferror(fq))
    {
        fprintf(stderr, "file error\n");
        return (EXIT_FAILURE);
    }

    fprintf(stderr, "processing complete\n");
    return (EXIT_SUCCESS);
}
