/*********************************************************************/
/*                                                                   */
/*  This Program Written by Paul Edwards.                            */
/*  Released to the Public Domain                                    */
/*                                                                   */
/*********************************************************************/
/*********************************************************************/
/*                                                                   */
/*  rdwcheck - analyze an RDW file (IEBCOPY unload)                  */
/*                                                                   */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static unsigned char rdw[4];
static unsigned char buf[70000];
static long offset = 0;
static long maxoff = 0;
static unsigned int maxlen = 0;
static unsigned int len;
static unsigned int ret;

int main(int argc, char **argv)
{
    FILE *fp;

    if (argc <= 1)
    {
        printf("usage: rdwcheck <rdw file>\n");
        return (EXIT_FAILURE);
    }

    fp = fopen(*(argv + 1), "rb");
    if (fp == NULL)
    {
        printf("can't open %s\n", *(argv + 1));
        return (EXIT_FAILURE);
    }

    while (1)
    {
        ret = fread(rdw, 1, 4, fp);
        if (ret == 0) break;
        if (ret < 4)
        {
            printf("only partial rdw at %ld\n", offset);
            break;
        }
        if ((rdw[2] != 0) || (rdw[3] != 0))
        {
            printf("last 2 bytes of RDW non-zero at %ld\n", offset);
            break;
        }
        len = rdw[0] << 8 | rdw[1];
        if (len <= 4)
        {
            printf("rdw too small (%d) at offset %ld\n", len, offset);
            break;
        }
        if (len > maxlen)
        {
            maxlen = len;
            maxoff = offset;
        }
        offset += 4;
        len -= 4;
        ret = fread(buf, 1, len, fp);
        if (ret != len)
        {
            printf("only read %u bytes instead of %u at offset %ld\n",
                   ret, len, offset);
            break;
        }
        offset += len;
    }
    printf("max RDW was %u at offset %ld\n", maxlen, maxoff);
    return (0);
}
