/*********************************************************************/
/*                                                                   */
/*  This Program Written by Paul Edwards.                            */
/*  Released to the Public Domain                                    */
/*                                                                   */
/*********************************************************************/
/*********************************************************************/
/*                                                                   */
/*  shiftsrt - shift a SRT file backwards or forwards in time        */
/*  00:00:59,640 --> 00:01:01,312                                    */
/*                                                                   */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

static char buf[1000];

static void updtime(char *x, long offset);
static double scale = 1.0;
static long firsttime = LONG_MIN;

int main(int argc, char **argv)
{
    FILE *fp;
    FILE *fq;
    char *t;
    char *t2;
    int off_secs;
    int off_msecs;
    long off_totmsecs;
    int state;

    if (argc <= 3)
    {
        printf("usage: shiftsrt <in file> <out file> <secs> [scale]\n");
        printf("e.g. shiftsrt in.srt out.srt -3.7 1.01\n");
        return (EXIT_FAILURE);
    }

    fp = fopen(*(argv + 1), "r");
    if (fp == NULL)
    {
        printf("can't open %s\n", *(argv + 1));
        return (EXIT_FAILURE);
    }

    fq = fopen(*(argv + 2), "w");
    if (fq == NULL)
    {
        printf("can't open %s\n", *(argv + 2));
        return (EXIT_FAILURE);
    }
    if (argc > 4)
    {
        scale = atof(*(argv + 4));
    }
    
    t = *(argv + 3);
    off_secs = atoi(t);
    off_totmsecs = off_secs * 1000;
    t2 = strchr(t, '.');
    if (t2 != NULL)
    {
        char bbb[] = "000";
        int l;

        t2++;
        if (strlen(t2) > 3)
        {
            printf("maximum number of digits after decimal point is 3\n");
            return (EXIT_FAILURE);
        }
        memcpy(bbb, t2, strlen(t2));
        off_msecs = atoi(bbb);
        if (off_secs < 0)
        {
            off_totmsecs -= off_msecs;
        }
        else
        {
            off_totmsecs += off_msecs;
        }
    }

    state = 0;
    while (fgets(buf, sizeof buf, fp) != NULL)
    {
        /* got the number */
        if ((state == 0) && (strlen(buf) > 1))
        {
            state = 1;
            fputs(buf, fq);
        }
        /* got the time */
        else if (state == 1)
        {
            state = 2;
            if (strlen(buf) < 27)
            {
                printf("malformed file\n");
                return (EXIT_FAILURE);
            }
            else if (memcmp(buf + 13, "-->", 3) != 0)
            {
                printf("malformed file 2\n");
                return (EXIT_FAILURE);
            }
            updtime(buf, off_totmsecs);
            updtime(buf + 17, off_totmsecs);
            fputs(buf, fq);
        }
        /* got the text */
        else if (state == 2)
        {
            state = 3;
            fputs(buf, fq);
        }
        /* another text line */
        else if ((state == 3) && (strlen(buf) > 1))
        {
            fputs(buf, fq);
        }
        /* got blank line */
        else if ((state == 3) && (strlen(buf) <= 1))
        {
            state = 0;
            fputs(buf, fq);
        }
    }
    fclose(fp);
    fclose(fq);
    printf("processing complete\n");
    return (0);
}

/* 00:00:59,640 --> 00:01:01,312 */
static void updtime(char *x, long offset)
{
    long oldtime;
    long newtime;
    int newhr;
    int newmin;
    int newsec;
    int newmsec;
    char y[20];

    oldtime = atoi(x) * 60;
    oldtime += atoi(x + 3);
    oldtime *= 60;
    oldtime += atoi(x + 6);
    oldtime *= 1000;
    oldtime += atoi(x + 9);
    
    if (firsttime == LONG_MIN)
    {
        firsttime = oldtime;
    }

    newtime = oldtime + offset;
    newtime = (double)newtime * scale;

    newmsec = newtime % 1000;
    newtime /= 1000;
    newsec = newtime % 60;
    newtime /= 60;
    newmin = newtime % 60;
    newtime /= 60;
    newhr = newtime;
    
    sprintf(y, "%02d:%02d:%02d,%03d", newhr, newmin, newsec, newmsec);
    memcpy(x, y, strlen(y));
    return;
}
